/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

/*! contacts.c
 *  This file contains the contact list handling functions.
 *  @author Bernhard Fischer <bf _at_ abenteuerland _dot_ at>
 *  @version 2014/06/16
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>

#include "log.h"
#include "sockaddr.h"
#include "contacts.h"



int lock_contacts(contact_list_t *cl)
{
#ifdef DEBUG_MX
   log_msg(LOG_DEBUG, "<contact list>");
#endif
   return pthread_mutex_lock(&cl->cl_mx);
}


int unlock_contacts(contact_list_t *cl)
{
#ifdef DEBUG_MX
   log_msg(LOG_DEBUG, "</contact list>");
#endif
   return pthread_mutex_unlock(&cl->cl_mx);
}


/*! Add a contact to a free slot in the contact list. This function does not
 * check for duplicates.
 * @param cl Pointer to the contact list.
 * @param fd File descriptor of contact.
 * @param ss Pointer socket address.
 * @return Returns the index number of the contact in the contact list or -1 in
 * case of error (no free slot in the list).
 */
int add_contact(contact_list_t *cl, int fd, const struct sockaddr *ss, const struct sockaddr *laddr, const char *name)
{
   int i;

   for (i = 0; i < MAX_CONTACTS; i++)
   {
      if (!cl->contact[i].fd)
      {
         log_msg(LOG_DEBUG, "adding contact %d on fd %d", i, fd);
         // clear all data (to be on the safe side in case if it is reused)
         memset(&cl->contact[i], 0, sizeof(cl->contact[i]));
         cl->contact[i].fd = fd;
         memcpy(&cl->contact[i].remote_addr, ss, SA_LEN(ss));
         cl->contact[i].local_addr = laddr;
         if (name != NULL)
         {
            strncpy(cl->contact[i].hostname, name, sizeof(cl->contact[i].hostname) - 1);
            // force \0-termination if name >= 256 bytes
            cl->contact[i].hostname[sizeof(cl->contact[i].hostname) - 1] = '\0';
         }
         else
            cl->contact[i].hostname[0] = '\0';
         break;
      }
   }

   if (i >= MAX_CONTACTS)
   {
      log_msg(LOG_INFO, "contact list full");
      //close(fd);
      return -1;
   }

   return i;
}

