/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

/*! cqueue.c
 *  This file contains the message queue functions for the message queue
 *  between the main thread and the connector thread.
 *  @author Bernhard Fischer <bf _at_ abenteuerland _dot_ at>
 *  @version 2014/06/16
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "cqueue.h"


/*! Free all resources used within queue control structure.
 * @param q Pointer to queue control structure.
 */
void queue_destroy(queue_t *q)
{
   pthread_mutex_destroy(&q->mx);
   close(q->cpipe[0]);
   close(q->cpipe[1]);
   free(q->queue);
}


/*! Initialize queue control structure.
 * @param q Pointer to queue control structure.
 * @return 0 on success, otherwise -1.
 */
int queue_init(queue_t *q)
{
   int err;

   q->queue = NULL;
   q->cnt = 0;
   if ((err = pthread_mutex_init(&q->mx, NULL)))
   {
      log_msg(LOG_ERR, "mutex_init() failed: %s", strerror(err));
      return -1;
   }

   if (pipe(q->cpipe) == -1)
   {
      log_errno(LOG_ERR, "cannot create queue pipe");
      pthread_mutex_destroy(&q->mx);
      return -1;
   }

   return 0;
}


/*! Unlock queue. 
 */
int queue_unlock(queue_t *q)
{
#ifdef DEBUG_MX
   log_msg(LOG_DEBUG, "</queue>");
#endif
   return pthread_mutex_unlock(&q->mx);
}


/*! Lock queue.
 */
int queue_lock(queue_t *q)
{
#ifdef DEBUG_MX
   log_msg(LOG_DEBUG, "<queue>");
#endif
   return pthread_mutex_lock(&q->mx);
}


int queue_wait(queue_t *q)
{
   char c;

   if (read(q->cpipe[0], &c, sizeof(c)) == -1)
      return -1;
   return c & 0xff;
}


/*! This function writes a character to the pipe. It is intendend to wake up a
 * blocking read on the other end. The reader could use the character to be
 * interpreted as commands.
 * @param q Pointer to queue control structure.
 * @param c Character to send.
 * @return This function directly returns the return value of write(2) (see
 * there) which may be -1, 0, and 1.
 */
int queue_signal_val(queue_t *q, char c)
{
   return write(q->cpipe[1], &c, sizeof(c));
}


/*! This function is identical to queue_signal_val(q, 0). See there for further
 * explaination.
 */
int queue_signal(queue_t *q)
{
   return queue_signal_val(q, 0);
}


/*! This function checks if an element is already in the queue.
 * @param q Pointer to queue control structure.
 * @param cmd Pointer to command which should be looked for in the queue.
 * @param cmp_func Comparison function which compares queue command elements.
 * The function shall return 0 if two elements are equal and non-0 if they are
 * not equal.
 * @return The function returns 1 if cmd is already queued otherwise 0.
 */
int queue_is_dup(queue_t *q, const command_t *cmd, int(*cmp_func)(const command_t*, const command_t*))
{
   int i;

   for (i = 0; i < q->cnt; i++)
      if (!cmp_func(cmd, &q->queue[i].command))
         return 1;
   return 0;
}


/*! This function adds a 'command' to the data queue. This function does not
 * lock the queue mutex.
 * @param q Pointer to the queue control structure.
 * @param cmd Command.
 * @return This function returns the number of elements on the queue after
 * queueing cmd or -1 on error (if realloc() failed).
 */
int queue_data_nomx(queue_t *q, command_t cmd)
{
   qmsg_t *qa;

   log_msg(LOG_DEBUG, "queuing data, q->cnt = %d", q->cnt);
   if ((qa = realloc(q->queue, sizeof(qmsg_t) * (q->cnt + 1))) == NULL)
   {
      log_errno(LOG_ERR, "realloc() failed");
      return -1;
   }

   q->queue = qa;
   q->queue[q->cnt].fd = 0;
   q->queue[q->cnt].status = MSG_NEW;
   q->queue[q->cnt].command = cmd;
   q->cnt++;
   queue_signal(q);
   
   return q->cnt;
}


/*! This function works exactly like queue_data_nomx() but locks the queue
 * mutex implicitly.
 */
int queue_data(queue_t *q, command_t cmd)
{
   int err;

   queue_lock(q);
   err = queue_data_nomx(q, cmd);
   queue_unlock(q);

   return err;
}


/*! Remove nth element from queue.
 * The queue must have been locked in advance.
 */
int queue_remove(queue_t *q, int n)
{
   // safety check for n
   if (n < 0 || n >= q->cnt)
   {
      log_msg(LOG_EMERG, "%d out of range!", n);
      return -1;
   }

   log_msg(LOG_DEBUG, "removing element %d on fd %d from queue, q->cnt = %d", n, q->queue[n].fd, q->cnt);
   q->cnt--;
   // move elements to index n of the array
   memmove(&q->queue[n], &q->queue[n + 1], sizeof(qmsg_t) * (q->cnt - n));

   return 0;
}


/*! Dequeue element from data queue.
 * @param q Pointer to queue control structure.
 * @param cmd Pointer to element which will receive the element of the queue.
 * @return Always returns 0.
 */
int dequeue_data(queue_t *q, command_t *cmd)
{
   for (;;)
   {
      queue_lock(q);
      if (q->cnt)
         break;
      queue_unlock(q);
      queue_wait(q);
   }

   *cmd = q->queue[0].command;
   queue_remove(q, 0);
   queue_unlock(q);

   log_msg(LOG_DEBUG, "data unqueued");
   return 0;
}

