/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

/*! dchat.c
 *  This file contains the main routines of Dchat.
 *  @author Bernhard Fischer <bf _at_ abenteuerland _dot_ at>
 *  @version 2014/06/16
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <signal.h>

#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifndef SOCK_NONBLOCK
#include <fcntl.h>
#endif

#include <pthread.h>


#include "log.h"
#include "dchat.h"
#include "message.h"
#include "sockaddr.h"
#include "contacts.h"
#include "socks.h"


static int max(int , int );
static int find_dup_name(const contact_list_t *, const char *);
static int find_dup(contact_list_t *, const struct sockaddr* );
static int frag_ = 0;
static volatile int sig_ = 0;


void sig_handler(int num)
{
   log_msg(LOG_DEBUG, "signal %d", num);
   sig_++;
}


/*! Create listening socket for incoming connections.
 * @param port Port number.
 * @return Returns a valid file descriptor or -1 in case of error.
 */
int init_server_socket(uint16_t port, int loop)
{
   struct sockaddr_storage sa;
   int s, so;

   memset(&sa, 0, sizeof(sa));
#ifndef AF_DEFAULT
#define AF_DEFAULT AF_INET
#endif
   sa.ss_family = AF_DEFAULT;
   sockaddr_setlen((struct sockaddr*) &sa);
   sockaddr_setport((struct sockaddr*) &sa, port);

   if (loop && sa.ss_family == AF_INET)
      ((struct sockaddr_in*) &sa)->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
   else if (loop && sa.ss_family == AF_INET6)
      ((struct sockaddr_in6*) &sa)->sin6_addr = in6addr_loopback;

   if ((s = socket(sa.ss_family, SOCK_STREAM, 0)) == -1)
   {
      log_errno(LOG_ERR, "socket() failed");
      return -1;
   }

   so = 1;
   if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &so, sizeof(so)) == -1)
      log_errno(LOG_ERR, "setsockopt() failed");

   if (bind(s, (struct sockaddr*) &sa, SA_LEN((struct sockaddr*) &sa)) == -1)
   {
      log_errno(LOG_ERR, "bind() failed");
      close(s);
      return -1;
   }

   if (listen(s, 10) == -1)
   {
      log_errno(LOG_ERR, "listen() failed");
      close(s);
      return -1;
   }

   return s;
}


/*! Check if sockaddr is in list of local addresses.
 * @return Returns the index of the duplicate address or -1 if it does not
 * exist.
 */
int check_addr(dchat_conf_t *cnf, const struct sockaddr *sa)
{
   int i;

   for (i = 0; i < cnf->addr_cnt; i++)
      if (!sockaddr_cmp(sa, (struct sockaddr*) &cnf->addr[i]))
         return i;
   return -1;
}


/*! Add sockaddr to list of local addresses.
 * @return Index of address in list. In case of error (list full) -1 is
 * returned.
 */
int add_addr(dchat_conf_t *cnf, const struct sockaddr *sa)
{
   if (cnf->addr_cnt >= MAX_IP)
      return -1;

   memcpy(&cnf->addr[cnf->addr_cnt], sa, SA_LEN(sa));
   cnf->addr_cnt++;

   return cnf->addr_cnt;
}


int msg_post(const message_t *msg, contact_t *ct)
{
   int len;

   if ((len = compose_raw_message(msg, ct->wbuf + ct->wlen, sizeof(ct->wbuf) - ct->wlen)) == -1)
      log_errno(LOG_ERR, "compose_raw_message() failed");
   else
      ct->wlen += len;

   return len;
}


void msg_init(const dchat_conf_t *cnf, message_t *msg)
{
   msg->port = cnf->port;
   if (cnf->hostname != NULL)
      strcpy(msg->hostname, cnf->hostname);
   else
      msg->hostname[0] = '\0';
}


void msg_init0(const dchat_conf_t *cnf, message_t *msg)
{
   memset(msg, 0, sizeof(*msg));
   msg_init(cnf, msg);
}


/*! This function prepares a new session (directly after accept() or connect())
 * for communication. The local address is stored to the list of local IP
 * addresses, a routing message is sent, and finally the session is added to
 * the contact list.
 * @param cnf Pointer to dchat config structure.
 * @param fd File descriptor of new session.
 * @param raddr Pointer to remote socket address (as supplied to connect() or
 * received from accept()).
 * @return The function returns the index of the new contact in the contact
 * list or -1 in case of error.
 */
int init_new_session(dchat_conf_t *cnf, int fd, struct sockaddr *raddr, const char *name)
{
   struct sockaddr_storage sa;
   struct sockaddr *laddr;
   socklen_t slen;
   message_t msg;
   char *rtxt;
   int n;

   // get local address
   log_msg(LOG_DEBUG, "get local address");
   slen = sizeof(sa);
   if (getsockname(fd, (struct sockaddr*) &sa, &slen) == -1)
   {
      log_errno(LOG_ERR, "getsockname() failed");
      return -1;
   }

   if (sockaddr_setport((struct sockaddr*) &sa, cnf->port) == -1)
   {
      // this should never happen
      log_msg(LOG_NOTICE, "address family not implemented");
      return -1;
   }

   // add to list of local addresses if it does not exist
   log_msg(LOG_DEBUG, "checking local address");
   pthread_mutex_lock(&cnf->mx);
   if ((n = check_addr(cnf, (struct sockaddr*) &sa)) == -1)
   {
      log_msg(LOG_DEBUG, "add address to local addresses");
      if ((n = add_addr(cnf, (struct sockaddr*) &sa)) == -1)
         log_msg(LOG_ERR, "list of local addresses is full");
   }
   pthread_mutex_unlock(&cnf->mx);

   // error checking of add_addr (outside locked segment)
   if (n == -1)
      return -1;

   // access to cnf->addr outside mutex is allowed, because list never shrinks!
   laddr = (struct sockaddr*) &cnf->addr[n];

   // create routing message
   log_msg(LOG_DEBUG, "creating routing message");
   msg_init0(cnf, &msg);
   lock_contacts(&cnf->cl);
   if ((rtxt = create_name_list(&cnf->cl)) == NULL)
   {
      unlock_contacts(&cnf->cl);
      log_msg(LOG_ERR, "create_name_list() failed");
      return -1;
   }
   mk_message(MSG_DISCOVER, rtxt, &msg);
   free(rtxt);
   if ((n = add_contact(&cnf->cl, fd, raddr, laddr, name)) != -1)
   {
      log_msg(LOG_DEBUG, "sending routing message");
      msg_post(&msg, &cnf->cl.contact[n]);
   }
   unlock_contacts(&cnf->cl);

   // add connection to contact list
   return n;
}


/*! Check socket for error (SO_ERROR).
 * @param fd File descriptor of socket.
 * @return Returns 0 if no error and -1 in case of error.
 */
static int check_socket(int fd)
{
   socklen_t err_len;
   int err;

   err_len = sizeof(err);
   if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &err_len) == -1)
   {
      log_errno(LOG_ERR, "getsockopt() failed");
      return -1;
   }
   if (err)
   {
      log_msg(LOG_NOTICE, "connection failed: %s", strerror(err));
      return -1;
   }
   return 0;
}


/*! This function asynchronously creates outgoing connections in parallel.
 */
void *multi_connector(dchat_conf_t *cnf)
{
   sigset_t ss;
   char c;
   int flags = 0;
   fd_set rset, wset;
   int nfds;
   int sig_exit = 0;
   int i, a;
   qmsg_t qmsg;

#ifdef HAVE_PTHREAD_GETNAME_NP
   pthread_setname_np(pthread_self(), "connector");
#endif
   // block all signals in this thread
   sigfillset(&ss);
   pthread_sigmask(SIG_BLOCK, &ss, NULL);
#ifdef SOCK_NONBLOCK
   flags = SOCK_NONBLOCK;
#endif

   while (!sig_exit)
   {
      // add signal pipe of queue to read set
      FD_ZERO(&rset);
      FD_SET(cnf->queue.cpipe[0], &rset);
      nfds = cnf->queue.cpipe[0];

      FD_ZERO(&wset);
      queue_lock(&cnf->queue);
      for (i = 0; i < cnf->queue.cnt; i++)
      {
         // add all connecting sockets to write set
         if (cnf->queue.queue[i].status == MSG_INPROG)
            FD_SET(cnf->queue.queue[i].fd, &wset);
         // add all connected sockets to read set for SOCKS handshake
         else if (cnf->queue.queue[i].status == MSG_READY)
            FD_SET(cnf->queue.queue[i].fd, &rset);
         else
            continue;

         nfds = max(nfds, cnf->queue.queue[i].fd);
      }
      queue_unlock(&cnf->queue);

      // wait for file descriptors
      if ((nfds = select(nfds + 1, &rset, &wset, NULL, NULL)) == -1)
      {
         log_errno(LOG_ERR, "select() failed in multi_connector");
         // FIXME: should probably handled different
         exit(1);
      }

      // check if new data is in queue
      if (FD_ISSET(cnf->queue.cpipe[0], &rset))
      {
         nfds--;
         sig_exit = queue_wait(&cnf->queue);

         // loop over all new elements (-> fd == 0) of the queue
         queue_lock(&cnf->queue);
         for (i = 0; i < cnf->queue.cnt; i++)
            if (!cnf->queue.queue[i].fd)  // -> status == MSG_NEW
            {
               // check if it is the own address
               pthread_mutex_lock(&cnf->mx);
               a = check_addr(cnf, (struct sockaddr*) &cnf->queue.queue[i].command);
               if (cnf->socks_addr.ss_family && !strcasecmp(cnf->hostname, cnf->queue.queue[i].command.hostname))
                  a = -2;
               pthread_mutex_unlock(&cnf->mx);
               if (a != -1)
               {
                  log_msg(LOG_NOTICE, "I do not connect to my-self");
                  queue_remove(&cnf->queue, i);
                  i--;
                  continue;
               }

               // create socket for new element
               if ((cnf->queue.queue[i].fd = socket(cnf->queue.queue[i].command.ss.ss_family, SOCK_STREAM | flags, 0)) == -1)
               {
                  log_errno(LOG_ERR, "could not create socket in multi_connector()");
                  queue_remove(&cnf->queue, i);
                  i--;
                  continue;
               }

#ifndef SOCK_NONBLOCK
               // manually set fd to non-blocking mode
               log_msg(LOG_DEBUG, "setting fd %d to O_NONBLOCK", cnf->queue.queue[i].fd);
               if ((fcntl(cnf->queue.queue[i].fd, F_SETFL, O_NONBLOCK)) == -1)
                  log_errno(LOG_ERR, "fcntl failed");
#endif
               // initiate connection
               log_sockaddr(LOG_DEBUG, (struct sockaddr*) &cnf->queue.queue[i].command);
               log_msg(LOG_DEBUG, "...connecting...");
               cnf->queue.queue[i].status = MSG_INPROG;
               if (connect(cnf->queue.queue[i].fd, (struct sockaddr*) &cnf->queue.queue[i].command.ss,
                        SA_LEN((struct sockaddr*) &cnf->queue.queue[i].command.ss)) == -1)
               {
                  if (errno == EINPROGRESS)
                  {
                     log_msg(LOG_DEBUG, "socket %d is trying to connect", cnf->queue.queue[i].fd);
                  }
                  else
                  {
                     // close and remove socket in case of error
                     log_errno(LOG_ERR, "connect() in multi_connector failed");
                     close(cnf->queue.queue[i].fd);
                     queue_remove(&cnf->queue, i);
                     i--;
                  }
                  continue;
               }
         
               // connection established immediately
               log_msg(LOG_DEBUG, "immediate connect()");
               FD_SET(cnf->queue.queue[i].fd, &wset);
               nfds++;
            }
         queue_unlock(&cnf->queue);
      } //if (FD_ISSET(cnf->queue.cpipe[0], &rset))

      // check if any connection was established
      while (nfds)
      {
         queue_lock(&cnf->queue);
         qmsg.fd = 0;
         for (i = 0; nfds && i < cnf->queue.cnt; i++)
         {
            if (cnf->queue.queue[i].fd && FD_ISSET(cnf->queue.queue[i].fd, &wset))
            {
               nfds--;
               // determine socket error
               if (check_socket(cnf->queue.queue[i].fd) == -1)
               {
                  close(cnf->queue.queue[i].fd);
                  queue_remove(&cnf->queue, i);
                  i--;
               }
               else
               {
                  log_msg(LOG_INFO, "connection established on fd = %d", cnf->queue.queue[i].fd);
                  // check if a SOCKS message should be sent
                  if (cnf->queue.queue[i].command.port)
                  {
                     cnf->queue.queue[i].status = MSG_READY;
                     if (socks_request(cnf->queue.queue[i].fd,
                              cnf->queue.queue[i].command.hostname, cnf->queue.queue[i].command.port) == -1)
                     {
                        log_errno(LOG_ERR, "onion_request failed");
                        close(cnf->queue.queue[i].fd);
                        queue_remove(&cnf->queue, i);
                        i--;
                     }
                  }
                  // direct connection (without SOCKS) ready
                  else
                  {
                     // create local copy
                     qmsg = cnf->queue.queue[i];
                  }
               }
            }
            else if (cnf->queue.queue[i].fd && FD_ISSET(cnf->queue.queue[i].fd, &rset))
            {
               nfds--;
               if (socks_response(cnf->queue.queue[i].fd) == SOCKS_GRANTED)
               {
                  // create local copy
                  qmsg = cnf->queue.queue[i];
               }
               else
               {
                  close(cnf->queue.queue[i].fd);
                  queue_remove(&cnf->queue, i);
                  i--;
               }
            }

            // if session is ready, break loop and handle outside queue_lock()
            if (qmsg.fd)
            {
               queue_remove(&cnf->queue, i);
               // handle new session outside of queue_lock to avoid deadlock
               break;
            }
         } // for (i = 0; nfds && i < cnf->queue.cnt; i++)
         queue_unlock(&cnf->queue);

         // if local copy contains valid data init connection
         if (qmsg.fd)
         {
            if ((i = init_new_session(cnf, qmsg.fd, (struct sockaddr*) &qmsg.command.ss, qmsg.command.hostname)) == -1)
            {
               log_msg(LOG_ERR, "init_new_session() failed, closing %d", qmsg.fd);
               close(qmsg.fd);
            }
            else
            {
               // remote port which is obviously is known in case of connect()
               lock_contacts(&cnf->cl);
               // but just if contact is still alive
               if (cnf->cl.contact[i].fd)
               {
                  cnf->cl.contact[i].port = sockaddr_port((struct sockaddr*) &qmsg.command.ss);
                  log_msg(LOG_DEBUG, "remote port set to %d", cnf->cl.contact[i].port);
               }
               unlock_contacts(&cnf->cl);

               // wake up select() in main_loop()
               c = 1;
               (void) write(cnf->cl_change[1], &c, 1);
            }
         }
      }
   } // while (!sig_exit)

   log_msg(LOG_DEBUG, "multi_connector() exiting");
   return NULL;
}


/*! Accept a new incoming TCP session on the listening socket and store remote
 * address to ss.
 * @param fd File descriptor if listening TCP socket.
 * @param ss Pointer to memory which will receive remote address.
 * @return Returns the file descriptor of the new TCP session or -1 in case of
 * error.
 */
int accept_once(int fd, struct sockaddr_storage *ss)
{
   socklen_t socklen;
   int s;

   socklen = sizeof(ss);
   if ((s = accept(fd, (struct sockaddr*) ss, &socklen)) == -1)
   {
      log_errno(LOG_ERR, "accept() failed");
      return -1;
   }

   log_msg(LOG_INFO, "incoming connection on fd %d", s);
   log_sockaddr(LOG_INFO, (struct sockaddr*) ss);

   return s;
}


/*! This function handles a new incoming TCP session, i.e. if the incoming
 * socket is ready for acceptance. The new session will be accepted and a
 * routing message will be sent to it immediately. Finally, the new contact is
 * added to the contact list.
 * @param cnf Pointer to dchat configuration structure.
 * @return Returns the number of the contact or -1 in case of error.
 */
int handle_tcp_incoming(dchat_conf_t *cnf)
{
   struct sockaddr_storage ss;
   int s, e;

   if ((s = accept_once(cnf->acpt_fd, &ss)) == -1)
   {
      log_msg(LOG_ERR, "accept_once() failed");
      return -1;
   }

   if ((e = init_new_session(cnf, s, (struct sockaddr*) &ss, NULL)) == -1)
   {
      log_msg(LOG_ERR, "init_new_session() failed, closing %d", s);
      close(s);
   }

   return e;
}


/*! Write the contents of the output buffer to the file descriptor of the
 * contact. The bytes written are removed from the output buffer. If the global
 * variable frag_ is 0 (which is default) all data is sent at once. If frag_ is
 * some positive value only the amount of frag_ is sent. Thus, if the buffer
 * contains more than frag_ bytes, write_once() shall be called repeatedly to
 * write all data.
 * @param ct Pointer to the contact.
 * @return The number of bytes actually written is returned or -1 in case of
 * error.
 */
int write_once(contact_t *ct)
{
   int len;

   if (!frag_ || ct->wlen < frag_)
      len = ct->wlen;
   else
      len = frag_;

   if ((len = write(ct->fd, ct->wbuf, len)) == -1)
   {
      log_errno(LOG_ERR, "could not write()");
      return -1;
   }

   memmove(ct->wbuf, ct->wbuf + len, ct->wlen - len);
   ct->wlen -= len;
   return len;
}


/*! Read data from 1 TCP session and write to local output.
 *  @param cnf Pointer to global config strucutre.
 *  @param n Number if tcp session (number of contact within list)
 *  @return ...
 */
int read_once(contact_t *ct)
{
   int len;

   if ((len = read(ct->fd, ct->rbuf + ct->rlen, /*sizeof(ct->buf)*/ FRAME_BUF_LEN - ct->rlen)) == -1)
   {
      log_msg(LOG_ERR, "read(%d) failed: '%s'", ct->fd, strerror(errno));
      return -1;
   }
   
   // handle eof
   if (!len)
   {
      log_msg(LOG_INFO, "contact on fd %d disconnected", ct->fd);
      return 0;
   }

   log_msg(LOG_DEBUG, "received %d bytes on fd %d", len, ct->fd);
   ct->rlen += len;
   return len;
}


/*! This function compares two command_t structures.
 * @return If both are equal, 0 is returned.
 */
int command_cmp(const command_t *a, const command_t *b)
{
   int sd;

   if ((sd = sockaddr_cmp((struct sockaddr*) &a->ss, (struct sockaddr*) &b->ss)))
      return sd;

   if (!a->port && !b->port)
      return 0;

   if ((sd = a->port - b->port))
      return sd;

   return strcmp(a->hostname, b->hostname);
}


int notify_connector(queue_t *q, const command_t *cmd)
{
   int n = 0;

   // request connection from connector
   queue_lock(q);
   if (!queue_is_dup(q, cmd, command_cmp))
   {
      queue_data_nomx(q, *cmd);
      n++;
   }
   else
      log_msg(LOG_NOTICE, "element already on queue");
   queue_unlock(q);

   return n;
}


/*! Parse command line commands of buffer.
 * @param cnf Pointer to DChat config structure.
 * @param str Pointer to \0-terminated string.
 * @return Returns number of command, or 0 if command unknown, or -1 in case of
 * error (e.g. syntax).
 */
int parse_console_input(dchat_conf_t *cnf, const char *str)
{
   char tmp[500], tmp0[200], buf[MAX_CONTENT_LENGTH + 1];
   command_t cmd;
   char *s, *endptr, *rtxt;
   int i, r, port = DEF_PORT;
   message_t msg;

   strcpy(buf, str);

   if ((s = strtok_r(buf, " \t\r\n", &endptr)) == NULL)
      return CMD_EMPTY;

   // check if it is a command (i.e. beginning with '/')
   if (*s != '/')
      return CMD_TEXT;

   if (!strcmp(s, "/connect"))
   {
      if ((s = strtok_r(NULL, " \t\r\n", &endptr)) == NULL)
         return CMD_E_SYNTAX;

      memset(&cmd, 0, sizeof(cmd));

      if (strlen(s) >= sizeof(cmd.hostname))
         return CMD_E_SYNTAX;

      // check if DChat shall used SOCKS
      if (cnf->socks_addr.ss_family)
      {
         // check if connection already exists
         if (find_dup_name(&cnf->cl, s) != -1)
         {
            log_msg(LOG_NOTICE, "already connected to '%s'", s);
            return CMD_CONNECT;
         }

         strcpy(cmd.hostname, s);
         // FIXME: default port shouldn't be hard coded
         cmd.port = DEF_PORT;
         // copy address of SOCKS server
         memcpy(&cmd.ss, &cnf->socks_addr, SA_LEN((struct sockaddr*) &cnf->socks_addr));
      }
      else
      {
         if (strchr(s, '.') != NULL)
         {
            cmd.ss.ss_family = AF_INET;
            r = inet_pton(AF_INET, s, &((struct sockaddr_in*) &cmd.ss)->sin_addr);
         }
         else if (strchr(s, ':') != NULL)
         {
            cmd.ss.ss_family = AF_INET6;
            r = inet_pton(AF_INET6, buf, &((struct sockaddr_in6*) &cmd.ss)->sin6_addr);
         }
         else
            return CMD_E_SYNTAX;

         // check return value of inet_pton()
         if (r != 1)
            return CMD_E_SYNTAX;

         sockaddr_setlen((struct sockaddr*) &cmd.ss);
         port = DEF_PORT;
         if ((s = strtok_r(NULL, " \t\r\n", &endptr)) != NULL)
         {
            if ((port = str_to_port(s)) == -1)
            {
               log_msg(LOG_ERR, "illegal port number, defaulting to %d", DEF_PORT);
               port = DEF_PORT;
            }
         }
         sockaddr_setport((struct sockaddr*) &cmd.ss, port);
      
         if (find_dup(&cnf->cl, (struct sockaddr*) &cmd.ss) != -1)
         {
            log_msg(LOG_NOTICE, "already connected to this peer");
            return CMD_CONNECT;
         }
      }

      notify_connector(&cnf->queue, &cmd);
      return CMD_CONNECT;
   }
   else if (!strcmp(s, "/help"))
   {
      r = snprintf(tmp, sizeof(tmp),
            "Commands:\n"
            "   /addresses ................ list of own (yet known) IP addresses\n"
            "   /close <contact> .......... close peer\n"
            "   /connect <ip> [<port>] .... connect to remote\n"
            "   /debug [on|off] ........... switch debug messages on/off\n"
            "   /discover ................. send routing message to all peers\n"
            "   /frag [0-65535] ........... set size of TCP segments\n"
            "   /help ..................... this help\n"
            "   /list ..................... show peers\n");
      (void) write(cnf->out_fd, tmp, r);
      return CMD_HELP;
   }
   else if (!strcmp(s, "/debug"))
   {
      if ((s = strtok_r(NULL, " \t\r\n", &endptr)) != NULL)
      {
         if (!strcmp(s, "off"))
            log_level(LOG_INFO);
         else if (!strcmp(s, "on"))
            log_level(LOG_DEBUG);
         else
            return CMD_E_SYNTAX;
      }
      else
         log_level(LOG_DEBUG);

      return CMD_DEBUG;
   }
   else if (!strcmp(s, "/frag"))
   {
      if ((s = strtok_r(NULL, " \t\r\n", &endptr)) != NULL)
      {
         if ((r = str_to_port(s)) != -1)
         {
            frag_ = r;
         }
         else
            return CMD_E_SYNTAX;
      }
      else
      {
         // set fragment size to 0 (i.e. no fragments)
         frag_ = 0;
      }

      return CMD_FRAG;
   }
   else if (!strcmp(s, "/addresses"))
   {
      pthread_mutex_lock(&cnf->mx);
      for (i = 0; i < cnf->addr_cnt; i++)
      {
         sockaddr_to_string(tmp0, sizeof(tmp0), (struct sockaddr*) &cnf->addr[i]);
         r = snprintf(tmp, sizeof(tmp), "%d: address = %s\n", i, tmp0);
         (void) write(cnf->out_fd, tmp, r);
      }
      pthread_mutex_unlock(&cnf->mx);

      return CMD_ADDRESSES;
   }
   else if (!strcmp(s, "/close"))
   {
      if ((s = strtok_r(NULL, " \t\r\n", &endptr)) == NULL)
         return CMD_E_SYNTAX;
      if ((r = str_to_port(s)) == -1)
         return CMD_E_SYNTAX;
      if (r >= MAX_CONTACTS)
         return CMD_E_SYNTAX;

      lock_contacts(&cnf->cl);
      if (!cnf->cl.contact[r].fd)
      {
         log_msg(LOG_ERR, "contact %d does not exist", r);
      }
      else
      {
         log_msg(LOG_INFO, "closing contact %d on fd %d", r, cnf->cl.contact[r].fd);
         close(cnf->cl.contact[r].fd);
         cnf->cl.contact[r].fd = 0;
      }
      unlock_contacts(&cnf->cl);

      return CMD_CLOSE;
   }
   else if (!strcmp(s, "/discover"))
   {
      // create routing message
      log_msg(LOG_DEBUG, "creating routing message");
      msg_init0(cnf, &msg);

      lock_contacts(&cnf->cl);
      if ((rtxt = create_name_list(&cnf->cl)) == NULL)
      {
         unlock_contacts(&cnf->cl);
         log_msg(LOG_ERR, "create_name_list() failed");
         return -1;
      }
      mk_message(MSG_DISCOVER, rtxt, &msg);
      free(rtxt);
      for (i = 0; i < MAX_CONTACTS; i++)
         if (cnf->cl.contact[i].fd)
         {
            // send message to new connection
            log_msg(LOG_DEBUG, "sending discover to contact %d on fd %d", i, cnf->cl.contact[i].fd);
            msg_post(&msg, &cnf->cl.contact[i]);
         }
      unlock_contacts(&cnf->cl);

      return CMD_DISCOVER;
   }
   else if (!strcmp(s, "/list"))
   {
      lock_contacts(&cnf->cl);
      for (i = 0; i < MAX_CONTACTS; i++)
         if (cnf->cl.contact[i].fd)
         {
            sockaddr_to_string(tmp0, sizeof(tmp0), (struct sockaddr*) &cnf->cl.contact[i].remote_addr);
            r = snprintf(tmp, sizeof(tmp), "%d: fd = %d, address = %s, remote port = %u, dir = %s, hostname = '%s'\n", i, cnf->cl.contact[i].fd, tmp0, cnf->cl.contact[i].port, cnf->cl.contact[i].port == sockaddr_port((struct sockaddr*) &cnf->cl.contact[i].remote_addr) ? "OUT" : "IN", cnf->cl.contact[i].hostname);
            (void) write(cnf->out_fd, tmp, r);
         }
      unlock_contacts(&cnf->cl);
      return CMD_LIST;
   }

   log_msg(LOG_WARN, "unknown command '%s'", s);
   return CMD_UNKNOWN;
}


/*! Output command prompt to console.
 */
static void prompt(const dchat_conf_t *cnf)
{
   char buf[100];

   snprintf(buf, sizeof(buf), "DCH/-\\T:%u> ", cnf->port);
   (void) write(cnf->out_fd, buf, strlen(buf));
}


/*! This function handles console input. It reads from the input file
 * descriptor (typically 0) and parse the command string. If it is a command
 * (such as e.g. '/connect') it is handled appropriately. If it is "just" text,
 * it is posted to all peers' output buffers which are connected and already
 * have a remote port number sent.
 * @param cnf Pointer to dchat config structure.
 * @return Returns the number of bytes read on input or -1 in case of error.
 */
int handle_console(dchat_conf_t *cnf)
{
   char buf[MAX_CONTENT_LENGTH + 1];
   message_t msg;
   int len, i, r;

   if ((len = read(cnf->in_fd, buf, sizeof(buf) - 1)) == -1)
   {
      log_msg(LOG_ERR, "read(%d) failed: '%s'", cnf->in_fd, strerror(errno));
      return -1;
   }

   if (!len)
   {
      log_msg(LOG_INFO, "fd %d EOF", cnf->in_fd);
      return 0;
   }

   buf[len] = '\0';
   if ((r = parse_console_input(cnf, buf)) < 0)
   {
      log_msg(LOG_WARN, "syntax error: %d", r);
      return -2;
   }
   if (r != CMD_TEXT)
   {
      //log_msg(LOG_INFO, "command %d executed", r);
      return -2;
   }

   // if the input contained no special command, send text message to all
   // contacts
   msg_init0(cnf, &msg);
   mk_text_message(buf, &msg);

   lock_contacts(&cnf->cl);
   for (i = 0; i < MAX_CONTACTS; i++)
   {
      if (cnf->cl.contact[i].fd && cnf->cl.contact[i].port)
      {
         if (msg_post(&msg, &cnf->cl.contact[i]) == -1)
            log_errno(LOG_ERR, "msg_post() failed");
      }
   }
   unlock_contacts(&cnf->cl);

   return len;
}


/*! Find contact with same hostname in contact list. The comparison is done
 * case-insensitive.
 * @param cl Pointer to contact list.
 * @param name Pointer to name which is looked up within cl.
 * @return Returns the index of the contact whose hostname matches name or -1
 * if no such contact was found.
 */
static int find_dup_name(const contact_list_t *cl, const char *name)
{
   int i;

   for (i = 0; i < MAX_CONTACTS; i++)
      // FIXME: this does not work yet for an unknown reason
      if (cl->contact[i].fd && !strcasecmp(cl->contact[i].hostname, name))
      {
         log_msg(LOG_DEBUG, "address is duplicate of contact %d on %d", i, cl->contact[i].fd);
         return i;
      }
   return -1;

}


/*! Find contact with the same listen port/IP as defined in socket address.
 * @param cl Pointer to contact list.
 * @param sa Socket address.
 * @return Returns index of contact which is a duplicate of sa or -1 if no
 * duplicate was found.
 */
static int find_dup(contact_list_t *cl, const struct sockaddr* sa)
{
   int i;

   for (i = 0; i < MAX_CONTACTS; i++)
      // FIXME: this does not work yet for an unknown reason
      if (cl->contact[i].fd && !sockaddr_cmp_addr((struct sockaddr*) &cl->contact[i].remote_addr, sa) &&
            sockaddr_port(sa) == cl->contact[i].port)
      {
         log_msg(LOG_DEBUG, "address is duplicate of contact %d on %d", i, cl->contact[i].fd);
         return i;
      }
   return -1;
}


/*! This function processes a discover message. It parses all addresses in the
 * list and connects to those which are not conected already.
 * @param cnf Pointer to dchat control structure.
 * @param msg Pointer to discover message.
 * @return The function returns the number of connections that where sent to
 * the connector or -1 in case of error.
 */
static int handle_discover(dchat_conf_t *cnf, const message_t *msg)
{
   struct sockaddr_storage *addr, *list;
   canon_addr_t *ca;
   command_t cmd;
   int i, n = 0;

   if ((ca = parse_canon_list(msg)) == NULL)
   {
      log_msg(LOG_ERR, "parse_canon_list() failed");
      return -1;
   }

   // if SOCKS is not in use...
   if (!cnf->socks_addr.ss_family)
   {
      list = parse_ip_list(ca);
      if (list == NULL)
      {
         log_msg(LOG_ERR, "parse_ip_list failed");
         free(ca);
         return -1;
      }

      for (addr = list; addr->ss_family != AF_UNSPEC; addr++)
         // check for duplicate address on list
         if (find_dup(&cnf->cl, (struct sockaddr*) addr) == -1)
         {
            memset(&cmd, 0, sizeof(cmd));
            memcpy(&cmd.ss, addr, SA_LEN((struct sockaddr*) addr));
            n += notify_connector(&cnf->queue, &cmd);
         }

      free(list);
   }
   else
   {
      for (i = 0; ca[i].name[0] != '\0'; i++)
      {
         if (find_dup_name(&cnf->cl, ca[i].name) == -1)
         {
            memset(&cmd, 0, sizeof(cmd));
            memcpy(&cmd.ss, &cnf->socks_addr, sizeof(cmd.ss));
            strcpy(cmd.hostname, ca[i].name);
            cmd.port = ca[i].port != -1 ? ca[i].port : DEF_PORT;
            n += notify_connector(&cnf->queue, &cmd);
         }
         else
         {
            log_msg(LOG_INFO, "name '%s' does already exist", ca[i].name);
         }
      }
   }

   free(ca);
   return n;
}

 
/*! This function reads data from a TCP socket if it is ready for reading into
 * the read buffer of the contact. It parse a message out of the buffer if a
 * full message is contained in the input buffer. Corrupt messages are silently
 * removed.
 * @param cnf Pointer to the DCHAT config structure.
 * @param ct Pointer to the contact which is ready for reading.
 * @return On success, the function returns the number of bytes that have been
 * successfully parsed out of the read buffer. The function returns 0 if the
 * buffer does not yet contain a complete message. Thus, it has to be called
 * again if the socket is ready for reading the next time. In case of a read
 * error -1 is returned.
 */
int handle_tcp_input(dchat_conf_t *cnf, contact_t *ct)
{
   message_t msg;
   char buf[200];
   int e, i;

   // read data into contact buffer
   switch (read_once(ct))
   {
      case -1:
         log_msg(LOG_ERR, "read_once() failed");
         return -1;
         
      case 0:
         log_msg(LOG_NOTICE, "contact closed session");
         return -1;
   }

   for (;;)
   {
      log_msg(LOG_DEBUG, "trying to parse message");
      // parse message and return if not enough data is in buffer
      if ((e = parse_message(ct, &msg)) == E_MSG_TRUNC)
      {
         //log_msg(LOG_DEBUG, "not enough data in buffer");
         return 0;
      }

      // in case of error resynchronize buffer
      if (e < 0)
      {
         log_msg(LOG_NOTICE, "resyncing buffer of fd %d", ct->fd);
         // return if no more data is in buffer
         if (!resync_buffer(ct))
            return 0;
         // restart parsing loop after resync
         continue;
      }

      // process message if everything is ok
      log_msg(LOG_DEBUG, "message type = %d, port = %u, len = %d", msg.type, msg.port, msg.len );

      // check if there's no port number in the contact list yet
      log_msg(LOG_DEBUG, "looking for duplicate session");
      for (i = 0; i < MAX_CONTACTS; i++)
         // if contact is active (fd != 0)
         if (cnf->cl.contact[i].fd &&
               // make sure to compare contact not to itself
               cnf->cl.contact[i].fd != ct->fd &&
               // and has a listen port
               cnf->cl.contact[i].port &&
               // and whose IP address is equal to this one's
               !sockaddr_cmp_addr((struct sockaddr*) &cnf->cl.contact[i].remote_addr, (struct sockaddr*) &ct->remote_addr) &&
               // and whose listen port is the same as this one announce
               cnf->cl.contact[i].port == msg.port)
         {
            // is duplicate, i.e. the same contact
            log_msg(LOG_DEBUG, "contact is duplicate of %d", i);
            // find out if we or the remote one has priority, i.e. the higher socket address
            e = sockaddr_cmp_addr(ct->local_addr, (struct sockaddr*) &ct->remote_addr);
            // compare listen ports if IP addresses are equal
            if (!e)
            {
               e = (int) cnf->port - msg.port;
               // self check
               if (!e)
               {
                  log_msg(LOG_ERR, "detected local loop");
                  return -1;
               }
            }
            
            // if out address is higher we have priority
            if (e > 0)
            {
               log_msg(LOG_DEBUG, "we have priority");
               // return -1 we cause this contact to be closed
               return -1;
            }
            
            // otherwise the other one has priority, thus we close the other contact
            log_msg(LOG_DEBUG, "the other one has priority");
            close(cnf->cl.contact[i].fd);
            cnf->cl.contact[i].fd = 0;
            break;
         }

      if (msg.port && ct->port != msg.port)
      {
         log_msg(LOG_NOTICE, "client changed port number to %u", msg.port);
         ct->port = msg.port;
      }

      if (msg.hostname[0] && strcasecmp(msg.hostname, ct->hostname))
      {
         log_msg(LOG_NOTICE, "client changed hostname to %s", msg.hostname);
         strcpy(ct->hostname, msg.hostname);
      }

      switch (msg.type)
      {
         case MSG_TEXT:
            sockaddr_to_string(buf, sizeof(buf), (struct sockaddr*) &ct->remote_addr);
            //(void) write(cnf->out_fd, "\n", 1);
            (void) write(cnf->out_fd, buf, strlen(buf));
            (void) write(cnf->out_fd, ": ", 2);
            (void) write(cnf->out_fd, msg.data, msg.len);
            (void) write(cnf->out_fd, "\n", 1);
            break;

         case MSG_DISCOVER:
            handle_discover(cnf, &msg);
            break;

         default:
            log_msg(LOG_NOTICE, "message type %d not implemented yet", msg.type);
      }
   }

   return e;
}


static int max(int a, int b)
{
   return a > b ? a: b;
}


int main_loop(dchat_conf_t *cnf)
{
   fd_set rset, wset;   // list of file descriptors for reading/writing
   int nfds;            // number of file descriptors in rset
   int i, n;
   char c;

   prompt(cnf);
   while (!sig_)
   {
      FD_ZERO(&rset);
      FD_ZERO(&wset);
      // add stdin to read-set
      FD_SET(cnf->in_fd, &rset);
      // add listener to read-set
      FD_SET(cnf->acpt_fd, &rset);
      nfds = max(cnf->in_fd, cnf->acpt_fd);
      // add signal pipe to read set
      FD_SET(cnf->cl_change[0], &rset);
      nfds = max(nfds, cnf->cl_change[0]);

      // add all tcp sessions to read-set
      lock_contacts(&cnf->cl);
      for (i = 0; i < MAX_CONTACTS; i++)
      {
         if (cnf->cl.contact[i].fd)
         {
            //log_msg(LOG_DEBUG, "adding contact %d, fd = %d to read set", i, cnf->cl.contact[i].fd);
            FD_SET(cnf->cl.contact[i].fd, &rset);
            // check if data is in output buffer
            if (cnf->cl.contact[i].wlen)
               FD_SET(cnf->cl.contact[i].fd, &wset);
            nfds = max(cnf->cl.contact[i].fd, nfds);
         }
      }
      unlock_contacts(&cnf->cl);

      if ((nfds = select(nfds + 1, &rset, &wset, NULL, NULL)) == -1)
      {
         if (errno == EINTR)
            continue;

         log_errno(LOG_ERR, "select() failed");
         // FIXME: improve error handling
         exit(1);
      }

      // console input is ready
      if (FD_ISSET(cnf->in_fd, &rset))
      {
         nfds--;
         if ((n = handle_console(cnf)) == -1)
            exit(1);
         if (!n)
            return 0;
         prompt(cnf);
      }

      // check signal pipe (from connector)
      if (FD_ISSET(cnf->cl_change[0], &rset))
      {
         nfds--;
         read(cnf->cl_change[0], &c, sizeof(c));
      }

      // check listening socket
      if (FD_ISSET(cnf->acpt_fd, &rset))
      {
         nfds--;
         (void) handle_tcp_incoming(cnf);
      }

      // check and handle input on tcp sessions
      lock_contacts(&cnf->cl);
      for (i = 0; nfds && i < MAX_CONTACTS; i++)
      {
         if (FD_ISSET(cnf->cl.contact[i].fd, &rset))
         {
            nfds--;
            if (handle_tcp_input(cnf, &cnf->cl.contact[i]) == -1)
            {
               log_msg(LOG_DEBUG, "closing contact %d on fd %d", i, cnf->cl.contact[i].fd);
               close(cnf->cl.contact[i].fd);
               cnf->cl.contact[i].fd = 0;
            }
         }
         if (FD_ISSET(cnf->cl.contact[i].fd, &wset))
         {
            nfds--;
            write_once(&cnf->cl.contact[i]);
         }
      }
      unlock_contacts(&cnf->cl);
   }

   log_msg(LOG_INFO, "^C was pressed");
   return 0;
}


void usage(const char *s)
{
   fprintf(stdout,
         "usage: %s [options]\n"
         "   -d ................ debug logging\n"
         "   -h ................ show this help\n"
         "   -p <port> ......... local listen port (default = %d)\n"
         "   -T <onion name> ... connect through Tor (SOCKS)\n",
         s, DEF_PORT);
}


int main(int argc, char **argv)
{
   struct sigaction sa;
   dchat_conf_t cnf;
   pthread_t conn_th;
   int e;

   memset(&cnf, 0, sizeof(cnf));
   cnf.in_fd = 0;
   cnf.out_fd = 1;
   cnf.port = DEF_PORT;

   while ((e = getopt(argc, argv, "dhp:T:")) != -1)
   {
      switch (e)
      {
         case 'd':
            log_level(LOG_DEBUG);
            break;

         case 'h':
            usage(argv[0]);
            exit(EXIT_SUCCESS);

         case 'p':
            if (!(cnf.port = atoi(optarg)))
               log_msg(LOG_ERR, "illegal argument, no port number"),
                  exit(EXIT_FAILURE);
            break;

         case 'T':
            cnf.hostname = optarg;
           break;
      }
   }

   if (cnf.hostname != NULL)
   {
      cnf.socks_addr.ss_family = AF_INET;
      sockaddr_setlen((struct sockaddr*) &cnf.socks_addr);
      sockaddr_setport((struct sockaddr*) &cnf.socks_addr, 9050);
      ((struct sockaddr_in*) &cnf.socks_addr)->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
   }
 
   memset(&sa, 0, sizeof(sa));
   sa.sa_handler = sig_handler;
   if ((sigaction(SIGINT, &sa, NULL)) == -1)
   {
      log_errno(LOG_ERR, "sigaction failed");
      exit(1);
   }

   if (queue_init(&cnf.queue) == -1)
   {
      log_msg(LOG_ERR, "queue_init() failed");
      exit(1);
   }

   if (pipe(cnf.cl_change) == -1)
   {
      log_errno(LOG_ERR, "could not create signal pipe");
      exit(1);
   }

   if ((e = pthread_mutex_init(&cnf.cl.cl_mx, NULL)))
   {
      log_msg(LOG_ERR, "mutex_init failed: %s", strerror(e));
      exit(1);
   }

   if ((e = pthread_mutex_init(&cnf.mx, NULL)))
   {
      log_msg(LOG_ERR, "mutex_init failed: %s", strerror(e));
      exit(1);
   }

   if ((cnf.acpt_fd = init_server_socket(cnf.port, cnf.hostname != NULL)) == -1)
      exit(1);

   if ((e = pthread_create(&conn_th, NULL, (void*(*)(void*)) multi_connector, &cnf)))
   {
      log_msg(LOG_ERR, "pthread_create() failed: %s", strerror(e));
      exit(1);
   }

   main_loop(&cnf);
   log_msg(LOG_DEBUG, "signalling connector for exit");
   queue_signal_val(&cnf.queue, 1);
   pthread_join(conn_th, NULL);

   queue_destroy(&cnf.queue);
   pthread_mutex_destroy(&cnf.cl.cl_mx);

   log_msg(LOG_INFO, "Good bye!");

   return 0;
}

