/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

/*! log.c
 *  This file contains the logging routines. They were originally introduced in
 *  2008 with the OnionCat project.
 *  @author Bernhard Fischer <bf _at_ abenteuerland _dot_ at>
 *  @version 2014/06/16
 */


#define _GNU_SOURCE

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include "log.h"


#ifndef LOG_PRI
#define LOG_PRI(p) ((p) & LOG_PRIMASK)
#endif

//! FILE pointer to log
static FILE *log_ = NULL;
//! log level
#ifdef DEBUG
static int level_ = LOG_DEBUG;
#else
static int level_ = LOG_INFO;
#endif
static const char *flty_[8] = {"emerg", "alert", "crit", "err", "warning", "notice", "info", "debug"};


static void __attribute__((constructor)) init_log0(void)
{
   log_ = stderr; 
}


void log_level(int lv)
{
   level_ = lv;
}


/*! Log a message to a file.
 *  @param out Open FILE pointer
 *  @param lf Logging priority (equal to syslog)
 *  @param fmt Format string
 *  @param ap Variable parameter list
 */
static void vlog_msgf(FILE *out, int lf, const char *fmt, va_list ap)
{
   static pthread_mutex_t mx = PTHREAD_MUTEX_INITIALIZER;
   int level = LOG_PRI(lf);
   char buf[1024];
   char name[20];

   if (level_ < level)
      return;

   name[0] = '\0';
#ifdef HAVE_PTHREAD_GETNAME_NP
   pthread_getname_np(pthread_self(), name, sizeof(name));
#else
   snprintf(name, sizeof(name), "%08x", (int) (long) pthread_self());
#endif
   pthread_mutex_lock(&mx);
   if (out != NULL)
   {
      fprintf(out, "[%7s] (%s) ", flty_[level], name);
      vfprintf(out, fmt, ap);
      fprintf(out, "\n");
   }
   else
   {
      vsnprintf(buf, sizeof(buf), fmt, ap);
      syslog(level | LOG_DAEMON, "(%s) %s", name, buf);
   }
   pthread_mutex_unlock(&mx);
}


/*! Log a message. This function automatically determines
 *  to which streams the message is logged.
 *  @param lf Log priority.
 *  @param fmt Format string.
 *  @param ... arguments
 */
void log_msg(int lf, const char *fmt, ...)
{
   va_list ap;

   va_start(ap, fmt);
   vlog_msgf(log_, lf, fmt, ap);
   va_end(ap);
}


void log_errno(int lf, const char *str)
{
   log_msg(lf, "%s: '%s'", str, strerror(errno));
}


/*! This function output len bytes starting at buf in hexadecimal numbers.
 *  */
void log_hex(int lf, const void *buf, int len)
{
   static const char hex[] = "0123456789abcdef";
   char  tbuf[100];
   int i;

   for (i = 0; i < len; i++)
   {
      snprintf(tbuf + (i & 0xf) * 3, sizeof(tbuf) - (i & 0xf) * 3, "%c%c ",
            hex[(((char*) buf)[i] >> 4) & 15], hex[((char*) buf)[i] & 15]);
      if ((i & 0xf) == 0xf)
         log_msg(lf, "%s", tbuf);
   }
   if ((i & 0xf))
     log_msg(lf, "%s", tbuf);
}

