/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

/*! message.c
 *  This file contains the Dchat protocol implementation.
 *  @author Bernhard Fischer <bf _at_ abenteuerland _dot_ at>
 *  @version 2014/06/16
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <arpa/inet.h>
#include <time.h>

#include "dchat.h"
#include "log.h"
#include "message.h"
#include "cqueue.h"
#include "sockaddr.h"


/*! Copy \n-terminated line from src to dst but not more that dst_len bytes
 * including a terminating \0. The function copies the \n-character as well.
 * dst will be \0-terminated by this function. If a line in src is longer than
 * dst_len or if src does not contain a terminating \n, dst will be truncated.
 * @param src Pointer to source data.
 * @param src_len Number of bytes in src.
 * @param dst Pointer to destination buffer.
 * @param dst_len Number of bytes in dst.
 * @return Returns the number of bytes copied, excluding the \0 character at
 * the end of dst.
 */
static int get_line(const char *src, int src_len, char *dst, int dst_len)
{
   int i;

   for (i = 0; i < src_len && i < dst_len - 1; i++, src++, dst++)
   {
      *dst = *src;
      if (*dst == '\n')
      {
         i++;
         dst++;
         break;
      }
   }
   *dst = '\0';

   return i;
}


/*! Split a message header of format "Key: Value" into parts. Finally, src will
 * contain the key only and the return value contains a pointer to the Value.
 * @return Returns a pointer to the value part of the string or NULL in case of
 * error.
 */
static char *split_header(char *src)
{
   char *endptr;
  
   if (strtok_r(src, ":", &endptr) == NULL)
      return NULL;

   return strtok_r(NULL, " \r\n", &endptr);
}


/*! This function resynchronizes the data buffer, i.e. it searches for the
 * occurance of the character 'D' and moves this to the beginning of the
 * buffer. If there is no 'D' the buffer is cleared.
 * @param ct Pointer to contact (which contains the buffer).
 * @return Returns the number of bytes remaining in the buffer.
 */
int resync_buffer(contact_t *ct)
{
   char *dst;

   // safety check: data in buffer?
   if (!ct->rlen)
      return 0;

   // find first occurrence of 'D'
   if ((dst = memchr(ct->rbuf, 'D', ct->rlen)) == NULL)
   {
      ct->rlen = 0;
      return 0;
   }

   // check if 'D' is already at the beginning of the buffer
   if (dst == ct->rbuf)
      // skip in that case
      dst++;

   memmove(ct->rbuf, dst, ct->rlen - (dst - ct->rbuf));
   ct->rlen -= dst - ct->rbuf;

   return ct->rlen;
}


/*! Parse message header into message_t structure.
 *  @param ct Pointer to contact.
 *  @param msg Pointer to message_t structure which will receive the message
 *  data.
 *  @return On success this function returns the number of bytes parsed in the
 *  input buffer (within ct). In case of error a negative number is returned
 *  which is defined in message.h as E_MSG_ macros. The msg structure may is
 *  undefined in case of error.
 */
int parse_message(contact_t *ct, message_t *msg)
{
   char buf[MAX_LINE_LENGTH + 1];
   int len, l, line;
   long clen;
   char *val;

   // init message structure
   memset(msg, 0, sizeof(*msg));

   // loop over all message headers
   for (line = 1, len = 0;; line++)
   {
      if (!(l = get_line(ct->rbuf + len, ct->rlen - len, buf, sizeof(buf))))
         return E_MSG_TRUNC;
      if (buf[l - 1] != '\n')
         return E_MSG_TRUNC;
      len += l;

      // check for termination line
      if (l == 1 || (l == 2 && buf[0] == '\r'))
         break;

      if ((val = split_header(buf)) == NULL)
         return E_MSG_HEAD1ILL;

      // treat 1st line special
      if (line == 1)
      {
         if (strcmp(buf, "DCHAT") || strcmp(val, "1.0"))
            return E_MSG_HEAD1ILL;
         continue;
      }

      // parse content type header
      if (!strcmp(buf, "Content-Type"))
      {
         if (!strcmp(val, "text/plain"))
            msg->type = MSG_TEXT;
         else if (!strcmp(val, "application/octet"))
            msg->type = MSG_OCTET;
         else if (!strcmp(val, "control/discover"))
            msg->type = MSG_DISCOVER;
         else
            log_msg(LOG_INFO, "unknown content type '%s'", buf);
      }
      else if (!strcmp(buf, "Listen-Port"))
      {
         errno = 0;
         clen = strtol(val, NULL, 0);
         if (errno)
         {
            log_errno(LOG_WARN, "illegal Listen-Port value");
            return E_MSG_PORTILL;
         }
         if (clen < 0 || clen >= 0x10000)
         {
            log_msg(LOG_WARN, "Listen-Port out of range");
            return E_MSG_PORTILL;
         }
         msg->port = clen;
      }
      // parse content length header
      else if (!strcmp(buf, "Content-Length"))
      {
         // convert string to long
         errno = 0;
         clen = strtol(val, NULL, 0);
         // error checking
         if (errno)
         {
            log_errno(LOG_WARN, "content length contains ill value");
            return E_MSG_LENILL;
         }
         if (clen > MAX_CONTENT_LENGTH)
         {
            log_msg(LOG_WARN, "content length exceeds maximum length: %ld", clen);
            return E_MSG_LENILL;
         }
         // everything ok
         msg->len = clen;
      }
      else if (!strcmp(buf, "Host"))
      {
         if (strlen(val) < sizeof(msg->hostname))
            strcpy(msg->hostname, val);
         else
            log_msg(LOG_WARN, "hostname of message too long!");
      }
      else
         log_msg(LOG_DEBUG, "ignoring header '%s'", buf);
   }

   if (!msg->type)
   {
      log_msg(LOG_WARN, "illegal header");
      return E_MSG_HEAD1ILL;
   }

   // check if enough data is in contact buffer
   if (ct->rlen - len < msg->len)
      return E_MSG_TRUNC;

   // copy payload to message structure
   memcpy(msg->data, ct->rbuf + len, msg->len);
   len += msg->len;
   // remove data from internal buffer
   memmove(ct->rbuf, ct->rbuf + len, ct->rlen - len);
   ct->rlen -= len;

   return len;
}


/*! Convert a charactor string to a valid port number.
 * @param s Pointer to string containing a number. The conversion is done with
 * strtol(3), thus it allows decimal, hexadecimal, and octal format.
 * @return It returns a port number in the range of 0 to 65535 or -1 in case of
 * error.
 */
int str_to_port(const char *s)
{
   long lval;

   errno = 0;
   lval = strtol(s, NULL, 0);
   if (errno)
   {
      log_msg(LOG_WARN, "malformed portnumber: %s", strerror(errno));
      return -1;
   }
   if (lval < 0 || lval >= 0x10000)
   {
      log_msg(LOG_WARN, "port out of range: %ld", lval);
      return -1;
   }

   return lval;
}


/*! Parse message body of formatted 2-column text in the form "<hostname>
 * [<port>]\r\n" into a list of canon_addr_t elements. <Hostname> may be any
 * string not longer than 255 characters. <Port> may be a valid portnumber in
 * the range of 1 to 65535. If the <port> field is missing, canon_addr_t.port
 * is set to -1. Lines with empty or illegal hostnames (strlen > 255) or
 * illegal port numbers are ignored.
 * @param msg Pointer to the message with the hostname/port list in its body.
 * @return The function returns a pointer to an array of type canon_addr_t. The
 * list is always terminated with a final canon_addr_t element which contains a
 * hostname with length 0. The memory is requested with a call to realloc(3),
 * thus, the caller of the function is responsible to free the memory again
 * with free(3). In case of error (i.e. realloc() fails), NULL is returned.
 */
canon_addr_t *parse_canon_list(const message_t *msg)
{
   canon_addr_t *tmp, *ca = NULL;
   char buf[MAX_LINE_LENGTH + 1];
   const char *src;
   char *s, *endptr;
   int len, l, n = 0, port;

   log_msg(LOG_DEBUG, "parse_ip_list()");
   for (src = msg->data, len = msg->len; len > 0; src += l, len -= l)
   {
      if ((l = get_line(src, len, buf, sizeof(buf))) <= 0)
         break;

      if ((s = strtok_r(buf, " \r\n", &endptr)) == NULL)
         continue;

      // safety check
      if (strlen(s) >= sizeof(ca[n].name))
      {
         log_msg(LOG_WARN, "name to long");
         continue;
      }

      n++;
      if ((tmp = realloc(ca, sizeof(*ca) * n)) == NULL)
      {
         log_errno(LOG_ERR, "realloc() failed");
         free(ca);
         return NULL;
      }
      ca = tmp;

      ca[n - 1].port = -1;
      strcpy(ca[n - 1].name, s);

      // parse port number if available
      if ((s = strtok_r(NULL, " \r\n", &endptr)) != NULL)
      {
         if (((port = str_to_port(s)) == -1) || !port)
         {
            log_msg(LOG_INFO, "ignoring illegal port number %d", port);
            continue;
         }
         ca[n - 1].port = port;
      }
   }

   // add NULL-element to the end of the list
   n++;
   if ((tmp = realloc(ca, sizeof(*ca) * n)) == NULL)
   {
      log_errno(LOG_ERR, "realloc() failed");
      free(ca);
      return NULL;
   }
   ca = tmp;
   ca[n - 1].port = -1;
   ca[n - 1].name[0] = '\0';

   return ca;
}


/*! Parse list of IPs/ports formatted as a canon_addr_t array into
 * an array of socket addresses. This function does not name lookups yet, thus
 * the names must be either IP or IPv6 addresses.
 * @param ca Pointer to list of names/ports.
 * @return A pointer to a socket address aray of type sockaddr_storage is
 * returned or NULL in case of error (see malloc(3) for errors). The list is
 * terminated with an element having the address family set to AF_UNSPEC. The
 * pointer returned by this function has to be freed by the caller with
 * free(3).
 */
struct sockaddr_storage *parse_ip_list(const canon_addr_t *ca)
{
   struct sockaddr_storage ss, *addr_list = NULL, *tmp;
   int i, n, r;

   log_msg(LOG_DEBUG, "parsing canon list");
   for (i = 0, n = 0; ca[i].name[0] != '\0'; i++)
   {
      r = 0;
      if (strchr(ca[i].name, '.') != NULL)
      {
          r = inet_pton(AF_INET, ca[i].name, &((struct sockaddr_in*) &ss)->sin_addr);
          ((struct sockaddr*) &ss)->sa_family = AF_INET;

      }
      // check for IPv6
      else if (strchr(ca[i].name, ':') != NULL)
      {
           r = inet_pton(AF_INET6, ca[i].name, &((struct sockaddr_in6*) &ss)->sin6_addr);
           ((struct sockaddr*) &ss)->sa_family = AF_INET6;
      }

      // ignore line
      if (r != 1)
      {
         log_msg(LOG_WARN, "client sent malformed IP address");
         continue;
      }

      // set port number in sockaddr structure
      sockaddr_setport((struct sockaddr*) &ss, ca[i].port != -1 ? ca[i].port : DEF_PORT);

      // add socket address to list
      n++;
      if ((tmp = realloc(addr_list, sizeof(*addr_list) * n)) == NULL)
      {
         log_errno(LOG_ERR, "realloc failed");
         free(addr_list);
         return NULL;
      }
      addr_list = tmp;
      memcpy(addr_list + n - 1, &ss, sizeof(*addr_list));
   }

   // add socket address to list
   n++;
   if ((tmp = realloc(addr_list, sizeof(*addr_list) * n)) == NULL)
   {
      log_errno(LOG_ERR, "realloc failed");
      free(addr_list);
      return NULL;
   }
   addr_list = tmp;
   addr_list[n - 1].ss_family = AF_UNSPEC;
 
   return addr_list;
}


/*! This function composes a raw message as sent on the wire out of a
 * structured message in msg.
 * @param msg Pointer to message_t structure.
 * @param buf Pointer to destination buffer which will receive the raw message.
 * @param size Size of the buffer in bytes.
 * @return The function returns the number of bytes written to buf. If either
 * msg or buf is NULL, the function will always return 0. If msg contains an
 * unknown message type, -1 is returned and errno set to EBADMSG. If there is
 * not enough space in buf -1 is returned and errno set to ENOBUFS.
 */
int compose_raw_message(const message_t *msg, char *buf, int size)
{
   char tmstr[100];
   char *typestr;
   int len;
   struct tm tm;
   time_t t;

   // safety check
   if (msg == NULL || buf == NULL || size <= 0)
      return 0;

   switch (msg->type)
   {
      case MSG_TEXT:
         typestr = "text/plain";
         break;

      case MSG_DISCOVER:
         typestr = "control/discover";
         break;

      default:
         errno = EBADMSG;
         return -1;
   }

   t = time(NULL);
   gmtime_r(&t, &tm);
   strftime(tmstr, sizeof(tmstr), "%a, %d %b %Y %T %z", &tm);
   len = snprintf(buf, size,
         "DCHAT: 1.0\r\n"
         "Content-Type: %s\r\n"
         "Date: %s\r\n"
         "Server: %s/%s\r\n",
         typestr, tmstr, PACKAGE_NAME, PACKAGE_VERSION);

   if (msg->port)
      len += snprintf(buf + len, size - len, "Listen-Port: %u\r\n", msg->port);

   if (msg->len)
      len += snprintf(buf + len, size - len, "Content-Length: %d\r\n", msg->len);

   if (msg->hostname[0])
      len += snprintf(buf + len, size - len, "Host: %s\r\n", msg->hostname);
      
   // add body separator
   len += snprintf(buf + len, size - len, "\r\n");

   if (size < len + msg->len)
   {
      errno = ENOBUFS;
      return -1;
   }

   memcpy(buf + len, msg->data, msg->len);
   return len + msg->len;
}


/*! This function fills out a message_t structure to be a text message.
 * @param text Pointer to \0-terminated text. If the length of the text exceeds
 * MAX_CONTENT_LENGTH it will be truncated.
 * @param msg Pointer to message_t structure to which the text is copied. The
 * message type is set accordingly (to MSG_TEXT).
 * @return The function returns the total number of bytes the structure
 * occupies in memory. Thus, it could be passed to e.g. malloc(). If text or
 * msg is NULL, mk_text_message() always returns 0.
 */
int mk_message(int type, const char *text, message_t *msg)
{
   // safety check
   if (text == NULL || msg == NULL)
      return 0;

   msg->len = strlen(text);
   if (msg->len > MAX_CONTENT_LENGTH)
      msg->len = MAX_CONTENT_LENGTH;

   msg->type = type;
   memcpy(msg->data, text, msg->len);

   return sizeof(*msg) - sizeof(msg->data) + msg->len;
}


int mk_text_message(const char *text, message_t *msg)
{
   return mk_message(MSG_TEXT, text, msg);
}


/*! This function creates a list of host addresses in the format: "<host>
 * <port>\r\n" for each contact. If the contact has a hostname it will be
 * filled in at <host>, otherwise the IP address will be filled out. It returns
 * a pointer to the list which was allocated with malloc(). Thus, it has to
 * freed again by the caller with free().
 * @param cl Pointer to contact list.
 * @return Returns a Pointer to the \0-terminated list or NULL in case of
 * error.
 */
char *create_name_list(const contact_list_t *cl)
{
#define IPV6STRLEN 39
   char *list, *tmp;
   int i, l, n, len, inc;

   if ((list = malloc(1)) == NULL)
   {
      log_errno(LOG_ERR, "malloc() failed");
      return NULL;
   }
   
   *list = '\0';
   len = 0;

   for (i = 0, n = 0; i < MAX_CONTACTS; i++)
      if (cl->contact[i].fd && cl->contact[i].port)
      {
         n++;
         inc = cl->contact[i].hostname[0] != '\0' ?  strlen(cl->contact[i].hostname) : IPV6STRLEN;
         inc += 9; // max. length of port number + ' ' + '\r\n' + '\0'

         if ((tmp = realloc(list, len + inc)) == NULL)
         {
            log_errno(LOG_ERR, "realloc() failed");
            free(list);
            return NULL;
         }
         list = tmp;

         if (cl->contact[i].hostname[0] != '\0')
         {
            len += snprintf(list + len, inc, "%s %u\r\n", cl->contact[i].hostname, cl->contact[i].port);
         }
         else
         {
            l = inet_ntop2(cl->contact[i].remote_addr.ss_family,
                  sockaddr_addrptr((struct sockaddr*) &cl->contact[i].remote_addr), list + len, inc);
            len += l;
            len += snprintf(list + len, inc - l, " %u\r\n", cl->contact[i].port);
         }
      }

   return list;
}

