/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

/*! sockaddr.c
 *  This file contains helper functions for manipulation of the socket address
 *  structures.
 *  @author Bernhard Fischer <bf _at_ abenteuerland _dot_ at>
 *  @version 2014/06/16
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "sockaddr.h"
#include "log.h"


/*! Return port number of sockaddr structure independently of its address
 * family.
 * @param sa Pointer to sockaddr structure.
 * @return Returns the port number in host byte order or -1 if address family
 * is not implemented.
 */
int sockaddr_port(const struct sockaddr *sa)
{
   if (sa->sa_family == AF_INET)
      return ntohs(((struct sockaddr_in*) sa)->sin_port);
   else if (sa->sa_family == AF_INET6)
      return ntohs(((struct sockaddr_in6*) sa)->sin6_port);
   return -1;
}


/*! Compare the address parts of two socket addresses (the IP addresses).
 * @return The function returns 0 if both addresses are equal, a positive value (1 to 255)
 * if the address of s1 is greater the s2 or a negative value (-1 to -255) if the address if
 * s1 is less than the address of s2. If the address families are different,
 * AF_INET6 is considered to be higher than AF_INET. If the family of s1 is
 * greater than the family of s2 SA_AF_GT is returned, otherwise SA_AF_LT. If
 * the address family of at least one of the two addresses is unknown
 * SA_ILL_FAMILY is returned.
 */
int sockaddr_cmp_addr(const struct sockaddr *s1, const struct sockaddr *s2)
{
   if (s1->sa_family == s2->sa_family)
   {
      if (s1->sa_family == AF_INET)
         return memcmp(&((struct sockaddr_in*) s1)->sin_addr, &((struct sockaddr_in*) s2)->sin_addr, 4);
      else if (s1->sa_family == AF_INET6)
         return memcmp(&((struct sockaddr_in6*) s1)->sin6_addr, &((struct sockaddr_in6*) s2)->sin6_addr, 16);
      return SA_ILL_FAMILY;
   }
   else if (s1->sa_family == AF_INET && s2->sa_family == AF_INET6)
      return SA_AF_LT;
   else if (s1->sa_family == AF_INET6 && s2->sa_family == AF_INET)
      return SA_AF_GT;
   return SA_ILL_FAMILY;
}


/*! Compare two sockaddr structures. This function uses sockaddr_cmp_addr()
 * first to compare the address family and the address. If they are equal it
 * further compares the port numbers.
 * @param s1 Pointer to sockaddr structure.
 * @param s2 Pointer to other sockaddr strucutre.
 * @return Returns 0 if they are completely equal. Values in the range of -255
 * to 255 indicate that the adresses are different.Values in the range if
 * -65535 << 8 and 65535 << 8 indicate that the addresses are equal but the
 *  port numbers differ.
 */
int sockaddr_cmp(const struct sockaddr* s1, const struct sockaddr *s2)
{
   int ret;

   if (!(ret = sockaddr_cmp_addr(s1, s2)))
      return (sockaddr_port(s1) - sockaddr_port(s2)) << 8;

   return ret;
}


/*! Convert socket address into character string of format "<address> <port>".
 * @param dst Pointer to destination buffer.
 * @param len Size of destination buffer.
 * @param sa Pointer to socket address structure.
 * @return Returns the length of the new string (see return value of
 * snprintf(3)).
 */
int sockaddr_to_string(char *dst, int len, const struct sockaddr *sa)
{
   char buf[100];
   int port;

   switch (sa->sa_family)
   {
      case AF_INET:
         inet_ntop(sa->sa_family, &((struct sockaddr_in*) sa)->sin_addr, buf,
               sizeof(buf));
         port = ntohs(((struct sockaddr_in*) sa)->sin_port);
         break;

      case AF_INET6:
         inet_ntop(sa->sa_family, &((struct sockaddr_in6*) sa)->sin6_addr, buf,
               sizeof(buf));
         port = ntohs(((struct sockaddr_in6*) sa)->sin6_port);
         break;
      default:
         log_msg(LOG_ERR, "invalid address family %d", sa->sa_family);
         port = 0;
         buf[0] = '\0';
   }
   //return snprintf(dst, len, "%d/%s/%d", sa->sa_family, buf, port);
   return snprintf(dst, len, "%s %d", buf, port);
}


/*! Set port number in socket address. This function takes care on the address
 * family, thus, the address family in sa must be set correctly.
 * @param sa Pointer to socket address.
 * @param port Port number in host byte order.
 * @return Returns 0 if everything is ok or -1 in case of error. That is that
 * the address family is not implemented or doesn't support port numbers.
 */
int sockaddr_setport(struct sockaddr *sa, uint16_t port)
{
   switch (sa->sa_family)
   {
      case AF_INET:
         ((struct sockaddr_in*) sa)->sin_port = htons(port);
         break;

      case AF_INET6:
         ((struct sockaddr_in6*) sa)->sin6_port = htons(port);
         break;

      default:
         return -1;
   }
   return 0;
}


/*! Set length field (e.g. sin_len) as it appears in BSD-style systems.
 * @param sa Pointer to socket address structure.
 * @return The function returns the length of the structure. If the address
 * family is unknown (or not implemented yet), -1 is returned. If sa is a
 * NULL-pointer 0 is returned.
 */
int sockaddr_setlen(struct sockaddr *sa)
{
   int len = 0;

   if (sa == NULL)
      return 0;

   if (!(len = SA_LEN(sa)))
      return -1;

#ifdef HAVE_SIN_LEN
   switch (sa->sa_family)
   {
      case AF_INET:
         ((struct sockaddr_in*) sa)->sin_len = len;
         break;

      case AF_INET6:
         ((struct sockaddr_in6*) sa)->sin6_len = len;
         break;
   }
#endif

   return len;
}


/*! Log socket address to log device using log_msg().
 * @param lf Log level (see syslog(3)).
 * @param sa Socket address.
 */
void log_sockaddr(int lf, const struct sockaddr *sa)
{
   char buf[100];

   sockaddr_to_string(buf, sizeof(buf), sa);
   log_msg(lf, "%s", buf);
}


/*! This function returns a pointer to the address field of a sockaddr
 * structure dependent on its address family.
 * @param sa Pointer to sockaddr structure.
 * @return Returns a pointer to the address within sa.
 */
const void *sockaddr_addrptr(const struct sockaddr *sa)
{
   switch (sa->sa_family)
   {
      case AF_INET:
         return &((struct sockaddr_in*) sa)->sin_addr;

      case AF_INET6:
         return &((struct sockaddr_in6*) sa)->sin6_addr;
   }
   return NULL;
}


/*! This function works exactly like inet_ntop(3) except that it returns the
 * number of bytes written to dst.
 * @param af Address family, currently only AF_INET and AF_INET6 are supported.
 * @param src Pointer to source address, e.g. struct in6_addr* or in_addr*.
 * @param dst Pointer to destination buffer.
 * @param size Number of bytes available in dst.
 * @return This function returns the number of bytes written to dst. In case of
 * error, -1 is returned and errno is set appropriately (see inet_ntop(3)). 
 */
int inet_ntop2(int af, const void *src, char *dst, socklen_t size)
{
   const char *s;

   if ((s = inet_ntop(af, src, dst, size)) == NULL)
      return -1;

   return strlen(s);
}

