/* Copyright 2014 Bernhard R. Fischer.
 *
 * This file is part of DChat.
 *
 * DChat is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * DChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DChat. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <netinet/in.h>

#include "socks.h"


#define SOCKS_USER "dchat"


/*! Send a SOCKS 4a request to a SOCKS server.
 *  @param fd File descriptor of connected session.
 *  @param hostname FQDN of destination host which shall be requested. This
 *  parameter  MUST NOT be NULL.
 *  @param port Destination port number.
 *  @return The function returns a positive value if the request was
 *  successfully sent. In case of error -1 is returned and errno is set
 *  appropriately (according to write(3)).
 */
static int socks_request0(int fd, const char *hostname, int port)
{
   char buf[sizeof(socks_hdr_t) + strlen(hostname) + 1 + strlen(SOCKS_USER) + 1];
   socks_hdr_t *shdr = (socks_hdr_t*) buf;

   shdr->ver = 4;
   shdr->cmd = 1;
   shdr->port = htons(port);
   shdr->addr.s_addr = htonl(1);
   memcpy(shdr + 1, SOCKS_USER, strlen(SOCKS_USER) + 1);
   memcpy((char*) (shdr + 1) + strlen(SOCKS_USER) + 1, hostname, strlen(hostname) + 1);

   return write(fd, buf, sizeof(*shdr) + strlen(SOCKS_USER) + 1 + strlen(hostname) + 1);
}


/*! This function directly calls socks_request0() if hostname is not NULL.
 *  @return If hostname is not NULL the return value of socks_request0() is
 *  returned. Otherwise -1 is returned and errno is set to EINVAL.
 */
int socks_request(int fd, const char *hostname, int port)
{
   // safety check
   if (hostname == NULL)
   {
      errno = EINVAL;
      return -1;
   }

   return socks_request0(fd, hostname, port);
}


/*! This function receives and interpretes the response from the SOCKS server.
 * @param fd File descriptor of session to SOCKS server.
 * @return The function returns the value of the status field of the SOCKS
 * response which is SOCKS_GRANTED (0x5a) if the request was successful. If
 * read() failed -1 is returned and errno set according to read(3). If the
 * response is truncated, -1 is returned and errno set to EBADMSG.
 */
int socks_response(int fd)
{
   socks_hdr_t shdr;
   int len;

   if ((len = read(fd, &shdr, sizeof(shdr))) == -1)
      return -1;

   if (len < sizeof(shdr))
   {
      errno = EBADMSG;
      return -1;
   }

   return shdr.cmd;
}


